//
// C++ Implementation: ncs36510updatefib
//
// Description: NCS36510 FIB dump and update utility
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>
//

#include "config.h"

#include "getopt.h"

#include <iostream>
#include <iomanip>
#include <sstream>
#include <stdexcept>
#include <cstring>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#ifdef HAVE_SYSEXITS_H
#include <sysexits.h>
#else
#define EX_USAGE   64
#define EX_DATAERR 65
#define EX_OK      0
#endif

#include <bfd.h>

class NCS36510FIB {
public:
	NCS36510FIB(void);
	~NCS36510FIB();
	bool open(const char *fn, bool writable);
	void close(void);
	std::ostream& dump(std::ostream& os);
	std::ostream& update(std::ostream& os);
	static std::ostream& dump_fib(std::ostream& os, const uint8_t *p, unsigned int size, unsigned int lma, const uint8_t *dflt);
	static std::ostream& dump_trim(std::ostream& os, const uint8_t *p, unsigned int size, unsigned int lma, const uint8_t *dflt);
	void calc_fib(uint8_t *p);
	void calc_trim(uint8_t *p);

protected:
	bfd *m_bfd;
	int m_fd;

	static std::ostream& dump_value(std::ostream& os, unsigned int sz, unsigned int offs, const uint8_t *p, const uint8_t *dflt);
};

NCS36510FIB::NCS36510FIB(void)
	: m_bfd(0)
{
}

NCS36510FIB::~NCS36510FIB()
{
	close();
}

void NCS36510FIB::close(void)
{
	if (m_bfd)
		bfd_close(m_bfd);
	else if (m_fd != -1)
		::close(m_fd);
	m_bfd = 0;
	m_fd = -1;
}

bool NCS36510FIB::open(const char *fn, bool writable)
{
	close();
	if (!fn) {
		std::cerr << "no file given" << std::endl;
		return false;
	}
	m_fd = ::open(fn, writable ? O_RDWR : O_RDONLY);
	if (m_fd == -1) {
		std::cerr << "cannot open file \"" << fn << "\"" << std::endl;
		return false;
	}
	m_bfd = bfd_fopen(fn, 0, "r", m_fd);
	if (!m_bfd) {
		::close(m_fd);
		m_fd = -1;
		std::cerr << "cannot open bfd file \"" << fn << "\"" << std::endl;
		return false;
	}
	if (!bfd_check_format(m_bfd, bfd_object)) {
		close();
		std::cerr << "file \"" << fn << "\" has invalid format" << std::endl;
		return false;
	}
}

std::ostream& NCS36510FIB::dump_value(std::ostream& os, unsigned int sz, unsigned int offs, const uint8_t *p, const uint8_t *dflt)
{
	os << std::hex;
	uint32_t val = 0;
	if (p) {
		for (unsigned int i = 0; i < sz; ++i)
			val |= p[i + offs] << (i * 8);
		os << " 0x" << std::setw(2 * sz) << std::setfill('0') << val;
	}
	if (dflt) {
		uint32_t vald = 0;
		for (unsigned int i = 0; i < sz; ++i)
			vald |= dflt[i + offs] << (i * 8);
		if (!p)
			os << "0x" << std::setw(2 * sz) << std::setfill('0') << vald << " (calculated)";
		else if (vald != val)
			os << " (expected 0x" << std::setw(2 * sz) << std::setfill('0') << vald << ')';
	}
	return os;
}

std::ostream& NCS36510FIB::dump_fib(std::ostream& os, const uint8_t *p, unsigned int size, unsigned int lma, const uint8_t *dflt)
{
	os << "NCS36510 FIB area";
	if (lma != 0x2000) {
		std::ostringstream oss;
		oss << std::hex << std::setw(8) << std::setfill('0') << lma;
		os << " (WARNING: nonstandard load address 0x" << oss.str() << ')';
	}
	os << std::endl;
	if (size < 40)
		return os << "  SIZE too short, " << size << " < 40" << std::endl;
	std::ostringstream oss;
	dump_value(oss << "  dummy_sp                ", 4, 0, p, dflt) << std::endl;
	dump_value(oss << "  dummy_reset_vector      ", 4, 4, p, dflt) << std::endl;
	dump_value(oss << "  dummy_nmi_handler       ", 4, 8, p, dflt) << std::endl;
	dump_value(oss << "  dummy_hardfault_handler ", 4, 12, p, dflt) << std::endl;
	dump_value(oss << "  dummy_blank             ", 4, 16, p, dflt) << std::endl;
	dump_value(oss << "  start                   ", 4, 20, p, dflt) << std::endl;
	dump_value(oss << "  size                    ", 4, 24, p, dflt) << std::endl;
	dump_value(oss << "  crc32                   ", 4, 28, p, dflt) << std::endl;
	dump_value(oss << "  fw_rev                  ", 4, 32, p, dflt) << std::endl;
	dump_value(oss << "  checksum                ", 4, 36, p, dflt) << std::endl;
	return os << oss.str();
}

std::ostream& NCS36510FIB::dump_trim(std::ostream& os, const uint8_t *p, unsigned int size, unsigned int lma, const uint8_t *dflt)
{
	os << "NCS36510 TRIM area";
	if (lma != 0x2800) {
		std::ostringstream oss;
		oss << std::hex << std::setw(8) << std::setfill('0') << lma;
		os << " (WARNING: nonstandard load address 0x" << oss.str() << ')';
	}
	os << std::endl;
	if (size < 24)
		return os << "  SIZE too short, " << size << " < 24" << std::endl;
	std::ostringstream oss;
	dump_value(oss << "  mac_addr_low            ", 4, 0, p, dflt) << std::endl;
	dump_value(oss << "  mac_addr_high           ", 4, 4, p, dflt) << std::endl;
	dump_value(oss << "  clk_32k_trim            ", 4, 8, p, dflt) << std::endl;
	dump_value(oss << "  clk_32m_trim            ", 4, 12, p, dflt) << std::endl;
	dump_value(oss << "  rssi                    ", 4, 16, p, dflt) << std::endl;
	dump_value(oss << "  txtune                  ", 4, 20, p, dflt) << std::endl;
	return os << oss.str();
}

std::ostream& NCS36510FIB::dump(std::ostream& os)
{
	if (!m_bfd)
		return os << "bfd not open" << std::endl;
	// section list
	if (true) {
		os << "Idx Name                 Size     VMA      LMA      Flags" << std::endl;
		unsigned int opb = bfd_octets_per_byte(m_bfd);
		for (asection *sec = m_bfd->sections; sec != NULL; sec = sec->next) {
			std::ostringstream oss;
			oss << std::setw(3) << sec->index << std::hex << ' '
			    << std::setw(20) << std::setiosflags(std::ios::left) << bfd_get_section_name(m_bfd, sec)
			    << std::resetiosflags(std::ios::left) << ' '
			    << std::setw(8) << std::setfill('0') << (bfd_section_size(m_bfd, sec) / opb) << ' '
			    << std::setw(8) << std::setfill('0') << bfd_get_section_vma(m_bfd, sec) << ' '
			    << std::setw(8) << std::setfill('0') << bfd_get_section_lma(m_bfd, sec) << ' '
			    << std::setw(8) << std::setfill('0') << sec->flags;
			if (sec->flags & SEC_HAS_CONTENTS)
				oss << " CONTENTS";
			if (sec->flags & SEC_ALLOC)
				oss << " ALLOC";
			if (sec->flags & SEC_CONSTRUCTOR)
				oss << " CONSTRUCTOR";
			if (sec->flags & SEC_LOAD)
				oss << " LOAD";
			if (sec->flags & SEC_RELOC)
				oss << " RELOC";
			if (sec->flags & SEC_READONLY)
				oss << " READONLY";
			if (sec->flags & SEC_CODE)
				oss << " CODE";
			if (sec->flags & SEC_DATA)
				oss << " DATA";
			if (sec->flags & SEC_ROM)
				oss << " ROM";
			if (sec->flags & SEC_DEBUGGING)
				oss << " DEBUGGING";
			if (sec->flags & SEC_NEVER_LOAD)
				oss << " NEVER_LOAD";
			if (sec->flags & SEC_EXCLUDE)
				oss << " EXCLUDE";
			if (sec->flags & SEC_SORT_ENTRIES)
				oss << " SORT_ENTRIES";
			if (bfd_get_arch(m_bfd) == bfd_arch_tic54x) {
				if (sec->flags & SEC_TIC54X_BLOCK)
					oss << " BLOCK";
				if (sec->flags & SEC_TIC54X_CLINK)
					oss << " CLINK";
			}
			if (sec->flags & SEC_SMALL_DATA)
				oss << " SMALL_DATA";
			switch (bfd_get_flavour(m_bfd)) {
			case bfd_target_coff_flavour:
				if (sec->flags & SEC_COFF_SHARED)
					oss << " SHARED";
				if (sec->flags & SEC_COFF_NOREAD)
					oss << " NOREAD";
				break;

			case bfd_target_elf_flavour:
				if (sec->flags & SEC_ELF_PURECODE)
					oss << " PURECODE";
				break;

			default:
				break;
			}
			if (sec->flags & SEC_THREAD_LOCAL)
				oss << " THREAD_LOCAL";
			if (sec->flags & SEC_GROUP)
				oss << " GROUP";
			if (bfd_get_arch(m_bfd) == bfd_arch_mep) {
				if (sec->flags & SEC_MEP_VLIW)
					oss << " VLIW";
			}
			os << oss.str() << std::endl;
		}
	}
	if (true) {
		unsigned int opb = bfd_octets_per_byte(m_bfd);
		for (asection *sec = m_bfd->sections; sec != NULL; sec = sec->next) {
			if (!(sec->flags & SEC_HAS_CONTENTS))
				continue;
			if (!(sec->flags & SEC_ALLOC))
				continue;
			if (!(sec->flags & SEC_LOAD))
				continue;
			bfd_size_type sz = bfd_section_size(m_bfd, sec);
			if (sz > 0x8000)
				continue;
			std::ostream& (*func)(std::ostream& os, const uint8_t *p, unsigned int size, unsigned int lma, const uint8_t *dflt) = 0;
			void (NCS36510FIB::*calc)(uint8_t *p) = 0;
			if (!strcmp(bfd_get_section_name(m_bfd, sec), ".fib")) {
				func = &NCS36510FIB::dump_fib;
				calc = &NCS36510FIB::calc_fib;
			} else if (!strcmp(bfd_get_section_name(m_bfd, sec), ".trim")) {
				func = &NCS36510FIB::dump_trim;
				calc = &NCS36510FIB::calc_trim;
			}
			if (!func)
				continue;
			uint8_t buf[sz];
			if (!bfd_get_section_contents(m_bfd, sec, buf, 0, sz))
				continue;
			uint8_t dflt[0x800];
			if (calc)
				(this->*calc)(dflt);
			func(os, buf, sz, bfd_get_section_lma(m_bfd, sec), calc ? dflt : 0);
		}
	}
	return os;
}

void NCS36510FIB::calc_fib(uint8_t *p)
{
	static const uint32_t dummy_sp                = 0x3ffffc00;
	static const uint32_t dummy_reset_vector      = 0x00003625;
	static const uint32_t dummy_nmi_handler       = 0x00003761;
	static const uint32_t dummy_hardfault_handler = 0x00003691;
	static const uint32_t dummy_blank             = 0x00000000;
	static const uint32_t fw_rev                  = 0x01000100;
	memset(p, 0xff, 0x800);
	p[0] = dummy_sp & 0xff;
	p[1] = (dummy_sp >> 8) & 0xff;
	p[2] = (dummy_sp >> 16) & 0xff;
	p[3] = (dummy_sp >> 24) & 0xff;
	p[4] = dummy_reset_vector & 0xff;
	p[5] = (dummy_reset_vector >> 8) & 0xff;
	p[6] = (dummy_reset_vector >> 16) & 0xff;
	p[7] = (dummy_reset_vector >> 24) & 0xff;
	p[8] = dummy_nmi_handler & 0xff;
	p[9] = (dummy_nmi_handler >> 8) & 0xff;
	p[10] = (dummy_nmi_handler >> 16) & 0xff;
	p[11] = (dummy_nmi_handler >> 24) & 0xff;
	p[12] = dummy_hardfault_handler & 0xff;
	p[13] = (dummy_hardfault_handler >> 8) & 0xff;
	p[14] = (dummy_hardfault_handler >> 16) & 0xff;
	p[15] = (dummy_hardfault_handler >> 24) & 0xff;
	p[16] = dummy_blank & 0xff;
	p[17] = (dummy_blank >> 8) & 0xff;
	p[18] = (dummy_blank >> 16) & 0xff;
	p[19] = (dummy_blank >> 24) & 0xff;
	p[32] = fw_rev & 0xff;
	p[33] = (fw_rev >> 8) & 0xff;
	p[34] = (fw_rev >> 16) & 0xff;
	p[35] = (fw_rev >> 24) & 0xff;
	uint32_t fibstart = -1, fibsize = -1, fibcrc32 = -1, fibchecksum = -1;
	static const uint32_t FLASHA_BASE = 0x00003000;
	static const uint32_t FLASHA_SIZE = 0x00052000;
	static const uint32_t FLASHB_BASE = 0x00102000;
	static const uint32_t FLASHB_SIZE = 0x00052000;
	uint8_t flash[FLASHA_SIZE + FLASHB_SIZE];
	memset(flash, 0xff, sizeof(flash));
	uint32_t starta = -1, enda = 0, startb = -1, endb = 0;
	for (asection *sec = m_bfd->sections; sec != NULL; sec = sec->next) {
		if (!strcmp(bfd_get_section_name(m_bfd, sec), ".fib"))
			continue;
		if (!strcmp(bfd_get_section_name(m_bfd, sec), ".trim"))
			continue;
		if (!(sec->flags & SEC_HAS_CONTENTS))
			continue;
		if (!(sec->flags & SEC_ALLOC))
			continue;
		if (!(sec->flags & SEC_LOAD))
			continue;
		bfd_size_type sz = bfd_section_size(m_bfd, sec);
		bfd_vma lma = bfd_get_section_lma(m_bfd, sec);
		if (lma >= FLASHA_BASE && lma + sz <= FLASHA_BASE + FLASHA_SIZE) {
			uint32_t addr = lma - FLASHA_BASE;
			uint32_t eaddr = addr + sz;
			if (!bfd_get_section_contents(m_bfd, sec, &flash[addr], 0, sz))
				std::cerr << "calc_fib: cannot load section " << bfd_get_section_name(m_bfd, sec) << std::endl;
			if (addr < starta)
				starta = addr;
			if (enda < eaddr)
				enda = eaddr;
			continue;
		}
		if (lma >= FLASHB_BASE && lma + sz <= FLASHB_BASE + FLASHB_SIZE) {
			uint32_t addr = lma - FLASHB_BASE;
			uint32_t eaddr = addr + sz;
			if (!bfd_get_section_contents(m_bfd, sec, &flash[FLASHA_SIZE + addr], 0, sz))
				std::cerr << "calc_fib: cannot load section " << bfd_get_section_name(m_bfd, sec) << std::endl;
			if (addr < startb)
				startb = addr;
			if (endb < eaddr)
				endb = eaddr;
			continue;
		}
		std::cerr << "calc_fib: section " << bfd_get_section_name(m_bfd, sec) << " outside FLASH" << std::endl;
	}
	if (enda <= starta) {
		std::cerr << "calc_fib: FLASHA empty" << std::endl;
		return;
	}
	if (endb <= startb) {
		fibstart = starta + FLASHA_BASE;
		fibsize = enda;
	} else {
		fibstart = starta + FLASHA_BASE;
		fibsize = endb + FLASHA_SIZE;
	}
	fibcrc32 = bfd_calc_gnu_debuglink_crc32(0, flash, fibsize);
	fibchecksum = (fibstart + fibsize + fibcrc32 + fw_rev) & 0xFFFFFFFF;
	p[20] = fibstart & 0xff;
	p[21] = (fibstart >> 8) & 0xff;
	p[22] = (fibstart >> 16) & 0xff;
	p[23] = (fibstart >> 24) & 0xff;
	p[24] = fibsize & 0xff;
	p[25] = (fibsize >> 8) & 0xff;
	p[26] = (fibsize >> 16) & 0xff;
	p[27] = (fibsize >> 24) & 0xff;
	p[28] = fibcrc32 & 0xff;
	p[29] = (fibcrc32 >> 8) & 0xff;
	p[30] = (fibcrc32 >> 16) & 0xff;
	p[31] = (fibcrc32 >> 24) & 0xff;
	p[36] = fibchecksum & 0xff;
	p[37] = (fibchecksum >> 8) & 0xff;
	p[38] = (fibchecksum >> 16) & 0xff;
	p[39] = (fibchecksum >> 24) & 0xff;
}

void NCS36510FIB::calc_trim(uint8_t *p)
{
        static const uint32_t mac_addr_low  = 0xffffffff;
        static const uint32_t mac_addr_high = 0xffffffff;
        static const uint32_t clk_32k_trim  = 0x00000039;
        static const uint32_t clk_32m_trim  = 0x00000017;
        static const uint32_t rssi          = 0x0000003d;
        static const uint32_t txtune        = 0xffffffff;
	memset(p, 0xff, 0x800);
	p[0] = mac_addr_low & 0xff;
	p[1] = (mac_addr_low >> 8) & 0xff;
	p[2] = (mac_addr_low >> 16) & 0xff;
	p[3] = (mac_addr_low >> 24) & 0xff;
	p[4] = mac_addr_high & 0xff;
	p[5] = (mac_addr_high >> 8) & 0xff;
	p[6] = (mac_addr_high >> 16) & 0xff;
	p[7] = (mac_addr_high >> 24) & 0xff;
	p[8] = clk_32k_trim & 0xff;
	p[9] = (clk_32k_trim >> 8) & 0xff;
	p[10] = (clk_32k_trim >> 16) & 0xff;
	p[11] = (clk_32k_trim >> 24) & 0xff;
	p[12] = clk_32m_trim & 0xff;
	p[13] = (clk_32m_trim >> 8) & 0xff;
	p[14] = (clk_32m_trim >> 16) & 0xff;
	p[15] = (clk_32m_trim >> 24) & 0xff;
	p[16] = rssi & 0xff;
	p[17] = (rssi >> 8) & 0xff;
	p[18] = (rssi >> 16) & 0xff;
	p[19] = (rssi >> 24) & 0xff;
	p[20] = txtune & 0xff;
	p[21] = (txtune >> 8) & 0xff;
	p[22] = (txtune >> 16) & 0xff;
	p[23] = (txtune >> 24) & 0xff;
	
}

std::ostream& NCS36510FIB::update(std::ostream& os)
{
	if (!m_bfd)
		return os << "bfd not open" << std::endl;
	uint8_t fib[0x800];
	calc_fib(fib);
	unsigned int upd(0);
	for (asection *sec = m_bfd->sections; sec != NULL; sec = sec->next) {
		if (strcmp(bfd_get_section_name(m_bfd, sec), ".fib"))
			continue;
		if (!(sec->flags & SEC_HAS_CONTENTS))
			continue;
		if (!(sec->flags & SEC_ALLOC))
			continue;
		if (!(sec->flags & SEC_LOAD))
			continue;
		bfd_size_type sz = bfd_section_size(m_bfd, sec);
		if (sz < 0x40)
			continue;
		if (sz > sizeof(fib))
			sz = sizeof(fib);
		if (lseek(m_fd, sec->filepos, SEEK_SET) == (off_t)-1) {
			os << "NCS36510FIB::update: seek error section " << bfd_get_section_name(m_bfd, sec)
			   << " size " << sz << std::endl;
			continue;
		}
		if (write(m_fd, fib, sz) != sz) {
			os << "NCS36510FIB::update: cannot update section " << bfd_get_section_name(m_bfd, sec)
			   << " size " << sz << std::endl;
			continue;
		}
		++upd;
	}
	if (!upd)
		return os << "NCS36510FIB::update: no .fib section found" << std::endl;
	{
		std::ostringstream oss;
		oss << std::hex
		    << "START 0x" << std::setw(8) << std::setfill('0')
		    << (fib[20] | (fib[21] << 8) | (fib[22] << 16) | (fib[23] << 24))
		    << " SIZE 0x" << std::setw(8) << std::setfill('0')
		    << (fib[24] | (fib[25] << 8) | (fib[26] << 16) | (fib[27] << 24))
		    << " CRC32 0x" << std::setw(8) << std::setfill('0')
		    << (fib[28] | (fib[29] << 8) | (fib[30] << 16) | (fib[31] << 24))
		    << " FW_REV 0x" << std::setw(8) << std::setfill('0')
		    << (fib[32] | (fib[33] << 8) | (fib[34] << 16) | (fib[35] << 24))
		    << " CHECKSUM 0x" << std::setw(8) << std::setfill('0')
		    << (fib[36] | (fib[37] << 8) | (fib[38] << 16) | (fib[39] << 24));
		os << oss.str() << std::endl;
	}
	return os;
}

int main(int argc, char *argv[])
{
	static const struct option long_options[] = {
		{ "dump", no_argument, 0, 'd' },
		{ "update", no_argument, 0, 'u' },
		{ "version", no_argument, 0, 'v' },
		{ 0, 0, 0, 0 }
	};
	int c, err = 0, mode = 0;
	while ((c = getopt_long(argc, argv, "duv", long_options, NULL)) != EOF) {
		switch (c) {
		case 'd':
			mode = 1;
			break;

		case 'u':
			mode = 2;
			break;

		case 'v':
			mode = 3;
			break;

		default:
			++err;
			break;
		}
	}
	if (err) {
		std::cerr << "usage: [-d] [-u] [-v]" << std::endl;
		return EX_DATAERR;
	}
	if (mode == 3) {
		std::cout << "ncs36510updatefib version " VERSION << std::endl;
		return EX_OK;
	}
	try {
		bfd_init();
		int ret = EX_OK;
		for (; optind < argc; ++optind) {
			NCS36510FIB fib;
			if (!fib.open(argv[optind], mode == 2)) {
				ret = EX_DATAERR;
				continue;
			}
			switch (mode) {
			case 1:
				fib.dump(std::cout);
				break;

			case 2:
				fib.update(std::cout);
				break;

			default:
				break;
			}
		}
		return ret;
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << std::endl;
		return EX_DATAERR;
	}
	return EX_OK;
}
